/*
    This file is part of gencon - a configuration file reader generator.
    Copyright (C) 2012 Sukru Cinar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lib.h"
#include "input.h"
#define FOR_REC(L,S) for(S=L;S;S=S->next)

static cstruct_t *reverse_structs(cstruct_t *head);
static void generate_header();
void print_types(FILE *out);
void print_struct(FILE *out, cstruct_t *S);
void print_field(FILE *out, cfield_t *F);
static void print_type(FILE *out, type_t *T);
static void generate_code();
static void print_cst_table(FILE *out, cstruct_t *S);
type_t *get_cst_type(type_t *T);
void print_cst(FILE *out,cst_t *C, type_t *T);
static void print_global_cst_tables(FILE *out);
static void print_top_level_cst(FILE *out, int kind, char *name);
void print_struct_table(FILE *out,cstruct_t *S);
void print_field_entry(FILE *out, cstruct_t *S, cfield_t *F);
static void print_top_function(FILE *out, cstruct_t *S,int body);
static FILE *open_file(char *fn);
static int field_kind(cfield_t *F);
static int print_readme_section(char *section);
char *parse_command_line(int argc,char **argv);

static int last_table;


static char* README=
"\n"
"The gencon program generates C code which reads a configuration file.\n"
"The configuration file can have nested sections and arrays of objects\n"
"in addition to three basic types: integers, floating point numbers\n"
"and strings. The generated code consists of a header file and a C source\n"
"file which needs no libraries other than libc. The header file defines\n"
"a struct for each section in your configuration file and appropriately typed\n"
"struct fields for each parameter. \n"
"\n"
"USAGE\n"
"\n"
" gencon -v                Prints the version number and exits.\n"
" gencon INPUT_FILE_NAME   Uses the given file as input.\n"
" gencon                   Uses stdin.\n"
" gencon -h                Prints this usage message.\n"
" gencon -H                Prints the whole README file.\n"
"\n"
"INPUT FILE\n"
"\n"
"gencon takes an input file and produces two C files: one header and one\n"
"source file. The input contains configuration file specifications and other\n"
"directives for options etc.\n"
"\n"
"A configuration file consists of parameter names and values. There\n"
"are three basic types for parameter values: integer, floating point\n"
"and string. Parameters can also be compound objects such as arrays\n"
"or collections of other parameters (similar to C structs). When a\n"
"configuration file is parsed, the contents of the whole configuration\n"
"file is stored in a C struct. gencon creates the definition of this\n"
"struct as well as the code to read it from a configuration file.\n"
"\n"
"This struct is defined by the \047top_level\047 directive in the input file:\n"
"\n"
"  top_level IDENT { FIELDs }\n"
"\n"
"The identifier for the top_level struct is used for both the type\n"
"of the generated struct and the corresponding read function. For example:\n"
"\n"
"  top_level settings { int a; }\n"
"\n"
"creates\n"
"\n"
"  typedef struct cfg_settings { int a; char G_a; } cfg_settings_t;\n"
"  char* cfg_parse_settings(char *file_name, cfg_settings_t *obj);\n"
"\n"
"The configuration file itself doesn\047t mention the name \047settings\047 at all.\n"
"It simply specifies parameter names and values. For example, the following\n"
"would be a valid configuration file to be parsed by cfg_parse_settings:\n"
"\n"
"  a 3\n"
"\n"
"As you can see, we have a single parameter for our example configuration\n"
"file, \047a\047 of the type \047int\047. When a parameter with a basic type is given,\n"
"gencon creates an extra field with type \047char\047 in the struct definition.\n"
"This field is then used as a \047given\047 indicator and is non-zero iff the\n"
"parsed configuration file specifies the parameter. The name of this field\n"
"is the same as the parameter it applies to but is prefixed with \042G_\042.\n"
"\n"
"In our example, the field specification of \047a\047 is exactly the same as\n"
"its C counterpart. This is accidental. Parameter declarations for gencon\n"
"have the following syntax\n"
"\n"
"  type  identifier  enumerations \047;\047\n"
"\n"
"\042enumerations\042 is optional and is explained later. \042type\042 can be one\n"
"of the basic types, an array or a struct. The other two basic types\n"
"are strings and floating point numbers. Strings are specified using the\n"
"keyword \047string\047 and the corresponding C type is \047char*\047. This pointer\n"
"is allocated using strdup(). Therefore, the null byte will terminate\n"
"the string if it\047s given. Otherwise, it appears at the end of the given\n"
"string. The floating point type is represented by the C type \047double\047\n"
"and is specified using the keyword \047float\047. To summarize the basic types,\n"
"we have the following:\n"
"\n"
"  int IDENT;\n"
"  float IDENT;\n"
"  string IDENT;\n"
"\n"
"Remember, basic-typed parameters have a G_* field, whereas compound ones\n"
"don\047t.\n"
"\n"
"Parameters can also be of array type. These are specified using the \047array\047\n"
"keyword. Arrays of arrays or multidimensional arrays are not supported. An\n"
"array can store any number of objects of the given type. The usage of the\n"
"array keyword is as follows:\n"
"\n"
"  array (int|float|string|STRUCT) IDENT enumerations;\n"
"\n"
"Here, IDENT is the name of the field and STRUCT is the name of a struct\n"
"defined previously. When an array parameter is specified, no G_* field\n"
"is generated but two fields specifying the size and length of the array\n"
"are:\n"
"\n"
" top_level array_example {\n"
"    array float foo;\n"
" }\n"
"\n"
"creates\n"
"\n"
" typedef struct cfg_array_example {\n"
"   double *foo; \n"
"   int N_foo, S_foo;\n"
" } cfg_array_example_t;\n"
"\n"
"Here, foo is allocated using malloc() and realloc(). N_foo holds the number\n"
"of elements in foo and S_foo holds the number of elements allocated. When\n"
"specifying values for an array in a configuration file, two methods may be\n"
"used. One of them is to specify elements one by one, mentioning the field\n"
"name in each case. For this example:\n"
"\n"
" foo 3.4\n"
" foo 6.7\n"
" foo -2.6e3\n"
"\n"
"will create a 3-element foo and store the values in there. Each time the\n"
"name of an array is mentioned, the given values are appended to the array.\n"
"The second method allows you to specify multiple elements at once:\n"
"\n"
"  foo { 1.2 3.5e-8 0.2 }\n"
"  foo { 0.5e9 0.0 }\n"
"\n"
"will create a 5- element \047foo\047. This second method can be used only if\n"
"the elements of the array are of basic types. It can\047t be used with \n"
"arrays of structs.\n"
"\n"
"Structs are compound objects which can be given as field types or as\n"
"array element types. Each struct must be defined before it\047s used\n"
"with the following directive:\n"
"\n"
" define IDENT { FIELDs }\n"
"\n"
"The semantics of this directive is exactly the same as the \047top_level\047\n"
"directive except for the fact that no read function is generated. When\n"
"you make such a declaration, gencon creates a struct declaration in the\n"
"header file:\n"
"\n"
" define font { float size; string family; }\n"
" top_level settings { font main_font; font message_font; }\n"
"\n"
"creates:\n"
"\n"
" typedef struct cfg_font { double size; char G_size;\n"
"                           char *family; char G_family; } cfg_font_t;\n"
" typedef struct cfg_settings {\n"
"    cfg_font_t main_font; cfg_font_t message_font; \n"
" } cfg_settings_t;\n"
"\n"
"As you can see, there is no G_* field for the font structures themselves,\n"
"but for their elements. When specifying values for struct parameters in a\n"
"configuration file, there are two methods similar to the array specification.\n"
"In the first one, you can give individual elements mentioning the struct\047s\n"
"name each time. This is just like the array single-element specifier. \n"
"\n"
"  main_font family \042Courier-New\042\n"
"  main_font size 13.5\n"
"\n"
"The second method is nicer, it reduces the amount of typing needed:\n"
"\n"
"  main_font { family \042Courier-New\042  size 13.5 }\n"
"\n"
"Also, the second method is somewhat mandatory if you have arrays of structs.\n"
"Let us have the following in addition to the definition of \047font\047 above:\n"
"\n"
"  top_level settings { array font styles; } \n"
"\n"
"When you specify values for \047styles\047, the left curly brace refers to the\n"
"newly created \047font\047 object rather than the \047styles\047 array. For instance:\n"
"\n"
"  styles { family \042Courier-New\042 size 13.5 }\n"
"\n"
"creates a new element in \047styles\047 with the given family and size. It doesn\047t\n"
"create two font objects, one with the given family and the other with the\n"
"given size. The following does exactly that\n"
"\n"
"  styles family \042Courier-New\042\n"
"  styles size 13.5\n"
"\n"
"which is equivalent to \n"
"\n"
"  styles { family \042Courier-New\042 }\n"
"  styles { size 13.5 }\n"
"\n"
"Therefore, it\047s impossible to use a multi-element specification for an array\n"
"if the array elements are struct-typed. i.e. the following doesn\047t work \n"
"at all.\n"
"\n"
"  styles {\n"
"   { family \042Courier-New\042 size 12.0 }\n"
"   { family \042Arial\042  size 10 }\n"
"  }\n"
"\n"
"This doesn\047t work, it probably will never work since it requires a token of\n"
"lookahead in the parser which I don\047t want to have just for this.\n"
"\n"
"This is it for the type specifications. gencon generated parsers can also\n"
"parse enumeration values if they are declared in the input file. For instance,\n"
"it makes more sense to write \047yes\047 or \047no\047 for a flag value instead of \0471\047 or\n"
"\0470\047. In such a case we would define an enumeration which maps \047yes\047 to \0471\047\n"
"and \047no\047 to \0470\047 for an integer parameter.\n"
"\n"
" top_level settings {\n"
"   int blork_code\n"
"       const\n"
"           yes (1)\n"
"           no  (0);\n"
"   int foo;\n"
" }\n"
"\n"
"The \047const\047 keyword introduces an enumeration of identifiers which is valid\n"
"only for the parameter it follows. For instance, a value of \047yes\047 would not\n"
"be valid for the \047foo\047 parameter above. A configuration file making use\n"
"of these enumerations could look like:\n"
"\n"
"  blork_code yes\n"
"  foo 55\n"
"\n"
"The constants must be given in parenthesis and must be compile-time constants\n"
"as they are put into global variable initializers. These can be any expression\n"
"yielding a constant. gencon doesn\047t try to interpret them in any way, it just\n"
"skips over quoted strings and balances the \047()\047 pairs.\n"
"\n"
"Enumerations can be given for only basic-typed parameters and array typed\n"
"parameters whose elements are basic-typed. It doesn\047t make sense to provide\n"
"an enumeration for a struct typed parameter. If you specify an enumeration\n"
"for an array parameter, then the enumeration is valid for the elements\n"
"of the array. For instance, if we have:\n"
"\n"
" top_level foo {\n"
"   array int blorzak\n"
"     const\n"
"       aha (\047X\047)\n"
"       ping (99)\n"
"       fifu (G_FIFU_VALUE);\n"
" }\n"
"\n"
"then the following would be a valid configuration file:\n"
"\n"
" blorzak { aha fifu ping fifu }\n"
" blorzak { 7 fifu ping 100 }\n"
"\n"
"It\047s also possible to specify \047global\047 constants which are available for all\n"
"fields defined in the input file with the same type. For this, you\047ll need\n"
"to mention the type of the enumeration values. Of course, only basic types\n"
"can be used for this purpose:\n"
"\n"
"  const int  IDENT (VALUE) IDENT (VALUE) .. ;\n"
"  const string  IDENT (VALUE) IDENT (VALUE) .. ;\n"
"  const float  IDENT (VALUE) IDENT (VALUE) .. ;\n"
"\n"
"The \047const\047 keyword can be used multiple times with the same type and has\n"
"a cumulative effect when you do so. For instance, if you have:\n"
"\n"
"  const int yes (1)  no (0);\n"
"  const int woman_maybe (0);\n"
"  top_level prefs {\n"
"     int error_messages;\n"
"     float threshold;\n"
"  }\n"
"\n"
"The enumerations are available for the \047error_messages\047 parameter but not\n"
"the \047threshold\047 field since the type of the enumeration is \047int\047. \n"
"\n"
"This is it for the configuration file specifications. \n"
"\n"
"DIRECTIVES\n"
"\n"
"There are also some directives you can use to tune the output.\n"
"\n"
"The directive \047%code_file\047 specifies the name of the C source file to be\n"
"generated, the default is \047config.c\047. It takes a string value:\n"
"\n"
" %code_file \042prefs.c\042\n"
"\n"
"Similarly, \047%header_file\047 specifies the name of the C header file.\n"
"The default is \047config.h\047.\n"
"\n"
" %header_file \042prefs.h\042\n"
"\n"
"If you want to include some code at the beginning or end of the generated\n"
"C source file, you can use the \047%header\047 and \047%footer\047 directives instead\n"
"of relying on \047cat\047.\n"
"\n"
" %header (\n"
"#include \042mylib.h\042\n"
" )\n"
"\n"
" %footer (\n"
"  cfg_config_t config;\n"
"  void read_all() {\n"
"    cfg_parse_config(\042.cfg\042, &config);\n"
"  }\n"
" )\n"
"As shown here, the code you want to put in the source file shall be enclosed\n"
"in balanced () pairs. This code isn\047t interpreted in any way, it should just\n"
"be legal regarding comments, quotes and \047()\047 pairs. \n"
"[[FIXME: do this for the header file as well. maybe rename header_file to\n"
"interface file and then add interface_header,footer and code_header, footer\n"
"etc.]]\n"
"\n"
"THE READ FUNCTIONS\n"
"\n"
"The generated read functions return an error string if there is an error, or\n"
"NULL if everything works OK. The first argument is used as the name of the \n"
"file to be read. \n"
"\n"
"The second argument is the object to be filled in. This object shall either\n"
"be allocated thru calloc() or shall be zeroed-out by memset() or global\n"
"declaration. If you fail to do so, array parameters will give you a hard\n"
"time.\n"
"\n"
"SOME TIPS\n"
"\n"
"Arrays of arrays are not supported, but can be worked around using structs\n"
"containing arrays and then having arrays of those structs. For instance:\n"
"\n"
" define row { array int values; }\n"
" top_level settings {\n"
"   array row rotation;\n"
" }\n"
"\n"
"would accept a configuration file such as:\n"
"\n"
" rotation values { 0 0 1 }\n"
" rotation values { 0 1 0 }\n"
" rotation values { 1 0 0 }\n"
"\n"
"There is no way to output to the standard output stream. Don\047t look for it.\n"
"\n"
"An input file can contain descriptions for multiple configuration file\n"
"formats, i.e. multiple \047top_level\047 declarations. If your program needs\n"
"to read multiple configuration files in different formats, then you need\n"
"to put all the descriptions in a single gencon input file and then run\n"
"gencon on it. The resulting code and header files will contain code to\n"
"read all the configuration file formats.  This is necessary because of\n"
"an implementation detail: flex and bison insist on generating \047extern\047\n"
"names instead of specifying an option to generate \047static\047 names.\n"
"\n"
"When you put multiple \047top_level\047 declarations in an input file, these\n"
"declarations share the \047define\047 declarations and global \047const\047 declarations.\n"
"\n"
"An input file can contain C-style block comments (/* comment */). Those\n"
"that are in C-texts will be copied to the output. The others simply get\n"
"removed from the input before anything else happens.\n"
"\n"
"In the output \047config.c\047 file, you don\047t need to include \047config.h\047 since\n"
"everything in \047config.h\047 is already printed in there also. This is for\n"
"allowing you to put config.h in some other directory if need be. For instance,\n"
"you could have\n"
"\n"
" %header_file \042../include/config.h\042\n"
"\n"
"and everything would still work since config.c doesn\047t need to\n"
"\n"
"  #include \042config.h\042\n"
"\n"
"CONFIGURATION FILES\n"
"\n"
"The output of gencon is some code which reads configuration files. This\n"
"section deals with some details about these files.\n"
"\n"
"A configuration file (CF for short) can contain comments using the #\n"
"style. Outside quoted strings, anything following a \047#\047 character is\n"
"ignored until the new next newline character. [[FIXME: should be valid\n"
"at EOF also.]]\n"
"\n"
"Values in a CF can be of three types: integers, strings and floating point\n"
"numbers. Integers can be either decimal or hexadecimal in C notation. A\n"
"leading 0 doesn\047t imply an octal number since octals are not recognized\n"
"at all. \n"
"\n"
"Floating point values are much simpler than their counterpart in C. A general\n"
"form for these values is:\n"
"\n"
" DIGITS . DIGITS exponent\n"
"\n"
"Both digit strings are mandatory and the exponent part is optional. i.e. the\n"
"following are not valid:\n"
"\n"
"  .2\n"
"  0.\n"
"  1.e3\n"
"\n"
"The exponent part is just like the C floating point exponent.\n"
"\n"
" (E|e)(+|-)DIGITS\n"
"\n"
"There is no support for hexadecimal floating point numbers. For both integers\n"
"and floating point numbers, you can include a leading minus sign to denote\n"
"negative numbers. \n"
"\n"
"String values must be quoted using either the single or double quote\n"
"characters. (ASCII 39 and 34, respectively.) Within the string, you can use\n"
"the following escape codes to encode special characters:\n"
"\n"
" \134a \134b \134f \134n \134r \134t \134v\n"
"   These have the same semantics as their C counterparts.\n"
" \134xHH\n"
"   This encodes a byte which corresponds to the two-digit hexadecimal\n"
"   value HH.\n"
" \134C\n"
"   This encodes the character C. Within a double quoted string, you don\047t\n"
"   need to escape a single quote and vice-versa.\n"
" No octal or Unicode escape sequences are recognized.\n"
"\n"
"INSTALLATION\n"
"\n"
"Just run \047make\047 in the source directory and put the resulting binary \n"
"\047gencon\047 somewhere in your $PATH. There are no files to install other\n"
"than that.\n"
"\n"
"I\047ve written this program on an Ubuntu 8.10 box with gcc 4.3.2 and it\n"
"works fine there. Let me know if you have a problem with the installation\n"
"or execution. See below for contact info.\n"
"\n"
"HACKING AND THINGS TO DO\n"
"\n"
"The distributed sources are not the original sources, but their processed\n"
"form. You can still hack on these, but contact me for the originals. I\047m\n"
"not distributing those because they are written to be used with another\n"
"tool (t2c) which requires some downloading and some installation which\n"
"I don\047t want to bother you with if you just want to install and use this.\n"
"\n"
"There are a couple of things to do, if you\047re interested:\n"
"\n"
"- The tool shall output line numbers (referring to the input file) in\n"
"  the generated source and header files. Ideally, it should switch back\n"
"  to the \047config.c\047 or \047config.h\047 file when some code from the library is\n"
"  inserted. For instance:\n"
"\n"
"  #line \042config.def\042 3\n"
"  typedef struct cfg_settings {\n"
"  #line \042config.def\042 11\n"
"     int a; char G_a;\n"
"  #line \042config.def\042 18\n"
"     int b; char G_b;\n"
"  } cfg_settings_t;\n"
"  #line \042config.h\042 6\n"
"  char *cfg_parse_settings(char *fn,cfg_settings_t *obj);\n"
"\n"
"  This would be very nice since we leave most sanity checks such as\n"
"  duplicate field/struct definitions or sanity check of the enumeration\n"
"  values to the compiler. It would be nice to have compiler error messages\n"
"  relating to the input file itself rather than the generated config.c/.h\n"
"  file.\n"
"\n"
"- A function to *write* a configuration object to a file should be \n"
"  generated as well. A GUI program can let the user change settings\n"
"  interactively and it would be nice to have an automatic writer.\n"
"\n"
"- We should be able to specify default values for parameters. These would\n"
"  be applied top-down. For instance, if you have:\n"
"\n"
"   define font { family \042Courier-New\042 size 14 }\n"
"   font main_font default { family \042Arial\042 };\n"
"\n"
"  Then, if the user doesn\047t specify a main_font, then it should end up with\n"
"  family \042Arial\042 and size 14. We could also nest these defaults arbitrarily.\n"
"  For instance:\n"
"\n"
"   style main_style default { font { family \042Arial\042 } color 0 };\n"
"\n"
"  When we have this, we must change the semantics of G_* flag fields also.\n"
"  Maybe I can have another flag such as D_* which indicates this. I don\047t\n"
"  think that size of a configuration object will be much of a problem.\n"
"\n"
"  For the default values, we could also allow C-text style values enclosed\n"
"  in \047()\047 pairs.\n"
"\n"
"- The values in enumerations shall be run-time values instead of compile-time\n"
"  values. i.e. foo() shall be eligible as an enumeration value. This would\n"
"  require me to put the values in lookup functions which are linked together\n"
"  by their last statement, chaining up to the the upper scope from an\n"
"  unsuccessful lookup in the current function.\n"
"\n"
"- Enumeration values could be scoped using \047{}\047 pairs. In the current setup,\n"
"  there are only two scopes for these: field scope and global scope. We could\n"
"  also allow struct-scope which isn\047t seen by other structs.\n"
"\n"
"- We could also implement some enum-only fields. These would be of any \n"
"  basic type but the value must be given as an identifier from a set of \n"
"  possibilities. This would represent \047choice\047 parameters better. However,\n"
"  we would have to implement reverse-lookup for the enumeration values since\n"
"  we want to write configuration files as well.\n"
"\n"
"- Regarding enumerations, we could implement a \047named enum\047. Then this\n"
"  named enum would be used instead of const specifications. For instance:\n"
"\n"
"   const int yesno\n"
"      yes(1) y(1) Y(1) \n"
"      no(0) n(0) N(0);\n"
"\n"
"   int blork_code yesno;\n"
"   int fifu_headers yesno;\n"
"\n"
"- The C99-style line comments are not recognized anywhere. (//comment).\n"
"\n"
"- There are some FIXMEs in this document which need attention.\n"
"\n"
"If you have any questions regarding this project, you can email me at this\n"
"address:\n"
"\n"
"  cinarus AT yahoo DOT com\n"
"\n"
"Please put the word \047gencon\047 somewhere in your subject line since I\n"
"filter out most email with subject \047Hello\047 or similar as spam.\n"
"\n"
"VERSION INFORMATION & LICENSING\n"
"\n"
"This is gencon-1.0. Copyright 2012 Sukru Cinar.\n"
"\n"
"    This program is free software: you can redistribute it and/or modify\n"
"    it under the terms of the GNU General Public License as published by\n"
"    the Free Software Foundation, either version 3 of the License, or\n"
"    (at your option) any later version.\n"
"\n"
"    This program is distributed in the hope that it will be useful,\n"
"    but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"    GNU General Public License for more details.\n"
"\n"
"    You should have received a copy of the GNU General Public License\n"
"    along with this program.  If not, see <http://www.gnu.org/licenses/>.\n"
"\n"
"CHANGE LIST\n"
"\n"
"20120226  First release gencon-1.0\n"

"\nQQ"
;


static cstruct_t *reverse_structs(cstruct_t *head)
{
  cstruct_t *newhead= 0, *S, *next;
  for(S=head;S;S=next)
  {
    next= S->next;
    S->next= newhead;
    newhead= S;
  }
  return newhead;
}
static void generate_header()
{
  FILE *out= open_file(header_file_name);
  cstruct_t *S;
  print_types(out);
  FOR_REC(top_level_structs,S) print_top_function(out,S,0);
  fclose(out);
}

void print_types(FILE *out)
{
  cstruct_t *S;
  FOR_REC(structs,S) print_struct(out,S);
  FOR_REC(top_level_structs,S) print_struct(out,S);
}


void print_struct(FILE *out, cstruct_t *S)
{
  cfield_t *F;
  fprintf(out,"typedef struct cfg_%s {\n", S->name);
  for(F=S->fields;F;F=F->next)
    print_field(out, F);
  fprintf(out,"} cfg_%s_t;\n", S->name);
}

void print_field(FILE *out, cfield_t *F)
{
  type_t *T;
  T= F->type;
  switch(T->kind) {
    case type_int: fprintf(out, "int %s; char G_%s;\n",
                     F->name, F->name); break;
    case type_string: fprintf(out, "char* %s; char G_%s;\n",
                     F->name, F->name); break;
    case type_float: fprintf(out, "double %s; char G_%s;\n",
                     F->name, F->name); break;
    case type_record: fprintf(out, "cfg_%s_t %s;\n", 
		     ((cstruct_t*) T->arg)->name, F->name); break;
    case type_array: print_type(out, T->arg);
                     fprintf(out," *%s; int N_%s; int S_%s;\n",
                           F->name, F->name, F->name); break;
  }
}


static void print_type(FILE *out, type_t *T)
{
  switch(T->kind)
   { case type_int: fprintf(out,"int"); break;
     case type_string: fprintf(out,"char*"); break;
     case type_float: fprintf(out,"double"); break;
     case type_record: fprintf(out, "cfg_%s_t",
                       ((cstruct_t*)T->arg)->name); break;
   }
}
static void generate_code()
{
   FILE *out= open_file(code_file_name);
   cstruct_t *S;
   fprintf(out,"%s", lib_header);
   print_types(out);
   FOR_REC(structs, S) print_cst_table(out, S);
   FOR_REC(top_level_structs, S) print_cst_table(out, S);
   print_global_cst_tables(out);
   FOR_REC(structs, S) print_struct_table(out,S);
   FOR_REC(top_level_structs, S) print_struct_table(out,S);
   fprintf(out,"%s", lib_footer);
   FOR_REC(top_level_structs, S) print_top_function(out,S,1);
   fclose(out);
}


static void print_cst_table(FILE *out, cstruct_t *S)
{
   cfield_t *F;
   for(F=S->fields;F;F=F->next)
   {
     cst_t *C;
     type_t *T;
     if (!F->cst) continue;
     F->table_number = ++last_table;
     fprintf(out,"static ");
     T= get_cst_type(F->type);
     switch(T->kind) {
      case type_int: fprintf(out,"cfgl_intcst_t"); break;
      case type_string: fprintf(out,"cfgl_strcst_t"); break;
      case type_float: fprintf(out,"cfgl_flocst_t"); break;
     }
     fprintf(out," cfgl_T%d[]= {\n", F->table_number);
     for(C=F->cst;C;C=C->next)
         print_cst(out,C, T);
     fprintf(out,"{0,0}\n");
     fprintf(out,"};\n");
   }
}


type_t *get_cst_type(type_t *T)
{
  if (T->kind!=type_array) return T;
  return T->arg;
}

void print_cst(FILE *out,cst_t *C, type_t *T)
{
  fprintf(out,"{ \"%s\", %s },\n", C->name, C->value);
}

static void print_global_cst_tables(FILE *out)
{
   print_top_level_cst(out, type_int, "int");
   print_top_level_cst(out, type_string, "str");
   print_top_level_cst(out, type_float, "flo");
}

static void print_top_level_cst(FILE *out, int kind, char *name)
{
  cst_t *C;
  fprintf(out,"static cfgl_%scst_t cfgl_global_%scst[]= {\n", name,name);
  for(C=top_level_consts;C;C=C->next)
    if (C->type->kind==kind)
      print_cst(out, C, C->type);
  fprintf(out," {0,0}\n");
  fprintf(out,"};\n");
}


void print_struct_table(FILE *out,cstruct_t *S)
{
  cfield_t *F;
  S->table_number= ++last_table;
  fprintf(out, "static cfgl_field_t cfgl_T%d[]={\n", S->table_number);
  for(F=S->fields;F;F=F->next)
     print_field_entry(out, S, F);
  fprintf(out," {0,0}\n");
  fprintf(out,"};\n");
}


void print_field_entry(FILE *out, cstruct_t *S, cfield_t *F)
{
  type_t *K, *T;
  int kind= field_kind(F);
  type_t *base_type;

  fprintf(out,"{");
  fprintf(out,"\"%s\", '%c' , ", F->name, kind);
  fprintf(out, "offsetof(cfg_%s_t,%s), ",  S->name, F->name);

  if (F->type->kind==type_array)
    fprintf(out,"offsetof(cfg_%s_t,S_%s), offsetof(cfg_%s_t,N_%s),",
                       S->name, F->name, S->name, F->name);
  else
    fprintf(out," 0,0,");

  if (F->type->kind==type_array) base_type= F->type->arg;
  else base_type= F->type;

  if (F->type->kind!=type_array && F->type->kind!=type_record)
      fprintf(out,"offsetof(cfg_%s_t,G_%s), ", S->name, F->name);
  else fprintf(out, "0, ");

  switch(base_type->kind) {
  case type_int: case type_string: case type_float:
     if (F->cst) fprintf(out," cfgl_T%d,", F->table_number);
     else fprintf(out, " 0, ");
     break;
  case type_record:
     fprintf(out," cfgl_T%d, ", ((cstruct_t*)base_type->arg)->table_number);
     break;
  case type_array:
     /* impossible */
     break;
  }

  if (kind=='R') 
     fprintf(out,"sizeof(cfg_%s_t)", ((cstruct_t*)base_type->arg)->name);
  else
     fprintf(out, "0"); 

  fprintf(out,"},\n");
}


static void print_top_function(FILE *out, cstruct_t *S,int body)
{
  fprintf(out, "char *cfg_parse_%s(char *fn, cfg_%s_t *obj)\n",
           S->name, S->name);
  if (body)
    fprintf(out,"{ return cfgl_parse_top(fn, obj, cfgl_T%d); }\n",
           S->table_number);
  else fprintf(out,";\n");
}

static FILE *open_file(char *fn)
{
  FILE *f= fopen(fn,"w");
  if (!f) exit(fprintf(stderr,"can't open %s\n", fn));
  return f;
}

static int field_kind(cfield_t *F)
{
  type_t *T= F->type;
  int isarray= 0;
  int R;
  if (T->kind==type_array) { isarray= 1; T= T->arg; }
  switch(T->kind) {
   case type_int: R= 'i'; break;
   case type_string: R= 's'; break;
   case type_float: R= 'f'; break;
   case type_record: R= 'r'; break;
  }
  if (isarray) return toupper(R);
  else return R;
}


static int print_readme_section(char *section)
{
  char *f,*s, *r= README;
again:
  f= strstr(r, section);
  if (!f) exit(fprintf(stderr,
                "internal error: can't find section '%s'\n",section)); 
  if (f[-1]!='\n') { r= f+1; goto again; }
  s= f;
  while(1) {
    f= strchr(f, '\n');
    if (!f) { f= README+strlen(README)-2; break; }
    if (isupper(f[1]) && isupper(f[2])) break;
    f++;
  }
  fwrite(s, 1, f-s, stdout);
  printf("\n");
  return 0;
}


char *parse_command_line(int argc,char **argv)
{
  int i;
  for(i=0;i<argc;i++)
   if (!strcmp(argv[i],"-v")) 
      exit(print_readme_section("VERSION"));
   else if (!strcmp(argv[i], "-h"))
      exit(print_readme_section("USAGE"));
   else if (!strcmp(argv[i],"-H")) {
      fwrite(README, 1, strlen(README)-2, stdout);
      exit(0);
   } 
   if (argc>2)
      exit(print_readme_section("USAGE")+1);
   if (argc==2)
     return argv[1];
   else return 0;
}


int main(int argc,char **argv)
{
  char *fn= parse_command_line(argc,argv);
  if (!fn) {
    parse_file(stdin, "(stdin)");
  } else {
    FILE *f= fopen(fn, "r");
    if (!f)
     exit(fprintf(stderr,"can't open %s\n", fn));
    parse_file(f, fn);
    fclose(f);
  }
  top_level_structs= reverse_structs(top_level_structs);
  structs= reverse_structs(structs);
  generate_header();
  generate_code();
  return 0;
}
